from ..timeoutException import TimeoutException
from .tradeException import TradeException
from ..errorHandler import ValidationException, NotFoundException, InternalException, UnauthorizedException
from .notSynchronizedException import NotSynchronizedException
from .notConnectedException import NotConnectedException
from .synchronizationListener import SynchronizationListener
from .reconnectListener import ReconnectListener
from ...metaApi.models import MetatraderHistoryOrders, MetatraderDeals, date, random_id, \
    MetatraderSymbolSpecification, MetatraderTradeResponse, MetatraderSymbolPrice, MetatraderAccountInformation, \
    MetatraderPosition, MetatraderOrder, format_date
from .latencyListener import LatencyListener
from .packetOrderer import PacketOrderer
from .packetLogger import PacketLogger
from .synchronizationThrottler import SynchronizationThrottler
import socketio
import asyncio
import re
from random import random
from datetime import datetime, timedelta
from typing import Coroutine, List, Dict


class MetaApiWebsocketClient:
    """MetaApi websocket API client (see https://metaapi.cloud/docs/client/websocket/overview/)"""

    def __init__(self, token: str, opts: Dict = None):
        """Inits MetaApi websocket API client instance.

        Args:
            token: Authorization token.
            opts: Websocket client options.
        """
        opts = opts or {}
        opts['packetOrderingTimeout'] = opts['packetOrderingTimeout'] if 'packetOrderingTimeout' in opts else 60
        self._application = opts['application'] if 'application' in opts else 'MetaApi'
        self._url = f'https://mt-client-api-v1.{opts["domain"] if "domain" in opts else "agiliumtrade.agiliumtrade.ai"}'
        self._request_timeout = opts['requestTimeout'] if 'requestTimeout' in opts else 60
        self._connect_timeout = opts['connectTimeout'] if 'connectTimeout' in opts else 60
        self._maxConcurrentSynchronizations = opts['maxConcurrentSynchronizations'] if 'maxConcurrentSynchronizations'\
                                                                                       in opts else 10
        retry_opts = opts['retryOpts'] if 'retryOpts' in opts else {}
        self._retries = retry_opts['retries'] if 'retries' in opts else 5
        self._minRetryDelayInSeconds = retry_opts['minDelayInSeconds'] if 'minDelayInSeconds' in retry_opts else 1
        self._maxRetryDelayInSeconds = retry_opts['maxDelayInSeconds'] if 'maxDelayInSeconds' in retry_opts else 30
        self._token = token
        self._requestResolves = {}
        self._synchronizationListeners = {}
        self._latencyListeners = []
        self._connected = False
        self._socket = None
        self._sessionId = None
        self._reconnectListeners = []
        self._connectedHosts = {}
        self._resubscriptionTriggerTimes = {}
        self._synchronizationThrottler = SynchronizationThrottler(self, self._maxConcurrentSynchronizations)
        self._synchronizationThrottler.start()
        self._packetOrderer = PacketOrderer(self, opts['packetOrderingTimeout'])
        if 'packetLogger' in opts and 'enabled' in opts['packetLogger'] and opts['packetLogger']['enabled']:
            self._packetLogger = PacketLogger(opts['packetLogger'])
            self._packetLogger.start()
        else:
            self._packetLogger = None

    async def on_out_of_order_packet(self, account_id: str, instance_index: int, expected_sequence_number: int,
                                     actual_sequence_number: int, packet: Dict, received_at: datetime):
        """Restarts the account synchronization process on an out of order packet.

        Args:
            account_id: Account id.
            instance_index: Instance index.
            expected_sequence_number: Expected s/n.
            actual_sequence_number: Actual s/n.
            packet: Packet data.
            received_at: Time the packet was received at.
        """
        print(f'[{datetime.now().isoformat()}] MetaApi websocket client received an out of order packet type ' +
              f'{packet["type"]} for account id {account_id}. Expected s/n {expected_sequence_number} does not ' +
              f'match the actual of {actual_sequence_number}')
        try:
            await self.subscribe(account_id, instance_index)
        except Exception as err:
            if err.__class__.__name__ != 'TimeoutException':
                print((f'[{datetime.now().isoformat()}] MetaApi websocket client failed to receive ' +
                       'subscribe response for account id ' + account_id + ':' + str(instance_index), err))

    def set_url(self, url: str):
        """Patch server URL for use in unit tests

        Args:
            url: Patched server URL.
        """
        self._url = url

    async def connect(self) -> asyncio.Future:
        """Connects to MetaApi server via socket.io protocol

        Returns:
            A coroutine which resolves when connection is established.
        """
        if not self._connected:
            self._connected = True
            self._requestResolves = {}
            self._resolved = False
            self._connectResult = asyncio.Future()
            self._packetOrderer.start()
            url = f'{self._url}?auth-token={self._token}'
            self._socket = socketio.AsyncClient(reconnection=False, request_timeout=self._request_timeout)

            while not self._socket.connected:
                try:
                    self._sessionId = random_id()
                    await asyncio.wait_for(
                        self._socket.connect(url, socketio_path='ws',
                                             headers={'Client-id': '{:01.10f}'.format(random())}),
                        timeout=self._connect_timeout)
                except Exception:
                    pass

            @self._socket.on('connect')
            async def on_connect():
                print(f'[{datetime.now().isoformat()}] MetaApi websocket client connected to the MetaApi server')
                if not self._resolved:
                    self._resolved = True
                    self._connectResult.set_result(None)

                if not self._connected:
                    self._socket.disconnect()

            @self._socket.on('connect_error')
            def on_connect_error(err):
                print(f'[{datetime.now().isoformat()}] MetaApi websocket client connection error', err)
                if not self._resolved:
                    self._resolved = True
                    self._connectResult.set_exception(Exception(err))

            @self._socket.on('connect_timeout')
            def on_connect_timeout(timeout):
                print(f'[{datetime.now().isoformat()}] MetaApi websocket client connection timeout')
                if not self._resolved:
                    self._resolved = True
                    self._connectResult.set_exception(TimeoutException('MetaApi websocket client connection timed out'))

            @self._socket.on('disconnect')
            async def on_disconnect():
                self._synchronizationThrottler.on_disconnect()
                print(f'[{datetime.now().isoformat()}] MetaApi websocket client disconnected from the MetaApi server')
                await self._reconnect()

            @self._socket.on('error')
            async def on_error(error):
                print(f'[{datetime.now().isoformat()}] MetaApi websocket client error', error)
                await self._reconnect()

            @self._socket.on('response')
            async def on_response(data):
                if data['requestId'] in self._requestResolves:
                    request_resolve = self._requestResolves[data['requestId']]
                    del self._requestResolves[data['requestId']]
                else:
                    request_resolve = asyncio.Future()
                self._convert_iso_time_to_date(data)
                if not request_resolve.done():
                    request_resolve.set_result(data)
                if 'timestamps' in data and hasattr(request_resolve, 'type'):
                    data['timestamps']['clientProcessingFinished'] = datetime.now()
                    for listener in self._latencyListeners:
                        try:
                            if request_resolve.type == 'trade':
                                await listener.on_trade(data['accountId'], data['timestamps'])
                            else:
                                await listener.on_response(data['accountId'], request_resolve.type, data['timestamps'])
                        except Exception as error:
                            print(f'[{datetime.now().isoformat()}] Failed to process on_response event for account ' +
                                  data['accountId'] + ', request type ' + request_resolve.type, error)

            @self._socket.on('processingError')
            def on_processing_error(data):
                if data['requestId'] in self._requestResolves:
                    request_resolve = self._requestResolves[data['requestId']]
                    del self._requestResolves[data['requestId']]
                    if not request_resolve.done():
                        request_resolve.set_exception(self._convert_error(data))

            @self._socket.on('synchronization')
            async def on_synchronization(data):
                if self._packetLogger:
                    self._packetLogger.log_packet(data)
                self._convert_iso_time_to_date(data)
                await self._process_synchronization_packet(data)

            return self._connectResult

    async def close(self):
        """Closes connection to MetaApi server"""
        if self._connected:
            self._connected = False
            await self._socket.disconnect()
            for request_resolve in self._requestResolves:
                if not self._requestResolves[request_resolve].done():
                    self._requestResolves[request_resolve].set_exception(Exception('MetaApi connection closed'))
            self._requestResolves = {}
            self._synchronizationListeners = {}
            self._latencyListeners = []
            self._reconnectListeners = []
            self._packetOrderer.stop()

    async def get_account_information(self, account_id: str) -> 'asyncio.Future[MetatraderAccountInformation]':
        """Returns account information for a specified MetaTrader account
        (see https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readAccountInformation/).

        Args:
            account_id: Id of the MetaTrader account to return information for.

        Returns:
            A coroutine resolving with account information.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getAccountInformation'})
        return response['accountInformation']

    async def get_positions(self, account_id: str) -> 'asyncio.Future[List[MetatraderPosition]]':
        """Returns positions for a specified MetaTrader account
        (see https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readPositions/).

        Args:
            account_id: Id of the MetaTrader account to return information for.

        Returns:
            A coroutine resolving with array of open positions.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getPositions'})
        return response['positions']

    async def get_position(self, account_id: str, position_id: str) -> 'asyncio.Future[MetatraderPosition]':
        """Returns specific position for a MetaTrader account
        (see https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readPosition/).

        Args:
            account_id: Id of the MetaTrader account to return information for.
            position_id: Position id.

        Returns:
            A coroutine resolving with MetaTrader position found.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getPosition',
                                                        'positionId': position_id})
        return response['position']

    async def get_orders(self, account_id: str) -> 'asyncio.Future[List[MetatraderOrder]]':
        """Returns open orders for a specified MetaTrader account
        (see https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readOrders/).

        Args:
            account_id: Id of the MetaTrader account to return information for.

        Returns:
            A coroutine resolving with open MetaTrader orders.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getOrders'})
        return response['orders']

    async def get_order(self, account_id: str, order_id: str) -> 'asyncio.Future[MetatraderOrder]':
        """Returns specific open order for a MetaTrader account
        (see https://metaapi.cloud/docs/client/websocket/api/readTradingTerminalState/readOrder/).

        Args:
            account_id: Id of the MetaTrader account to return information for.
            order_id: Order id (ticket number).

        Returns:
            A coroutine resolving with metatrader order found.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getOrder', 'orderId': order_id})
        return response['order']

    async def get_history_orders_by_ticket(self, account_id: str, ticket: str) -> MetatraderHistoryOrders:
        """Returns the history of completed orders for a specific ticket number
        (see https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readHistoryOrdersByTicket/).

        Args:
            account_id: Id of the MetaTrader account to return information for.
            ticket: Ticket number (order id).

        Returns:
            A coroutine resolving with request results containing history orders found.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getHistoryOrdersByTicket',
                                                        'ticket': ticket})
        return {
            'historyOrders': response['historyOrders'],
            'synchronizing': response['synchronizing']
        }

    async def get_history_orders_by_position(self, account_id: str, position_id: str) -> MetatraderHistoryOrders:
        """Returns the history of completed orders for a specific position id
        (see https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readHistoryOrdersByPosition/)

        Args:
            account_id: Id of the MetaTrader account to return information for.
            position_id: Position id.

        Returns:
            A coroutine resolving with request results containing history orders found.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getHistoryOrdersByPosition',
                                                        'positionId': position_id})
        return {
            'historyOrders': response['historyOrders'],
            'synchronizing': response['synchronizing']
        }

    async def get_history_orders_by_time_range(self, account_id: str, start_time: datetime, end_time: datetime,
                                               offset=0, limit=1000) -> MetatraderHistoryOrders:
        """Returns the history of completed orders for a specific time range
        (see https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readHistoryOrdersByTimeRange/)

        Args:
            account_id: Id of the MetaTrader account to return information for.
            start_time: Start of time range, inclusive.
            end_time: End of time range, exclusive.
            offset: Pagination offset, default is 0.
            limit: Pagination limit, default is 1000.

        Returns:
            A coroutine resolving with request results containing history orders found.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getHistoryOrdersByTimeRange',
                                                        'startTime': format_date(start_time),
                                                        'endTime': format_date(end_time),
                                                        'offset': offset, 'limit': limit})
        return {
            'historyOrders': response['historyOrders'],
            'synchronizing': response['synchronizing']
        }

    async def get_deals_by_ticket(self, account_id: str, ticket: str) -> MetatraderDeals:
        """Returns history deals with a specific ticket number
        (see https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readDealsByTicket/).

        Args:
            account_id: Id of the MetaTrader account to return information for.
            ticket: Ticket number (deal id for MT5 or order id for MT4).

        Returns:
            A coroutine resolving with request results containing deals found.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getDealsByTicket',
                                                        'ticket': ticket})
        return {
            'deals': response['deals'],
            'synchronizing': response['synchronizing']
        }

    async def get_deals_by_position(self, account_id: str, position_id: str) -> MetatraderDeals:
        """Returns history deals for a specific position id
        (see https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readDealsByPosition/).

        Args:
            account_id: Id of the MetaTrader account to return information for.
            position_id: Position id.

        Returns:
            A coroutine resolving with request results containing deals found.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getDealsByPosition',
                                                        'positionId': position_id})
        return {
            'deals': response['deals'],
            'synchronizing': response['synchronizing']
        }

    async def get_deals_by_time_range(self, account_id: str, start_time: datetime, end_time: datetime, offset: int = 0,
                                      limit: int = 1000) -> MetatraderDeals:
        """Returns history deals with for a specific time range
        (see https://metaapi.cloud/docs/client/websocket/api/retrieveHistoricalData/readDealsByTimeRange/).

        Args:
            account_id: Id of the MetaTrader account to return information for.
            start_time: Start of time range, inclusive.
            end_time: End of time range, exclusive.
            offset: Pagination offset, default is 0.
            limit: Pagination limit, default is 1000.

        Returns:
            A coroutine resolving with request results containing deals found.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getDealsByTimeRange',
                                                        'startTime': format_date(start_time),
                                                        'endTime': format_date(end_time),
                                                        'offset': offset, 'limit': limit})
        return {
            'deals': response['deals'],
            'synchronizing': response['synchronizing']
        }

    def remove_history(self, account_id: str, application: str = None) -> Coroutine:
        """Clears the order and transaction history of a specified application so that it can be synchronized from
        scratch (see https://metaapi.cloud/docs/client/websocket/api/removeHistory/).

        Args:
            account_id: Id of the MetaTrader account to remove history for.
            application: Application to remove history for.

        Returns:
            A coroutine resolving when the history is cleared.
        """
        params = {'type': 'removeHistory'}
        if application:
            params['application'] = application
        return self._rpc_request(account_id, params)

    def remove_application(self, account_id: str) -> Coroutine:
        """Clears the order and transaction history of a specified application and removes the application
        (see https://metaapi.cloud/docs/client/websocket/api/removeApplication/).

        Args:
            account_id: Id of the MetaTrader account to remove history and application for.

        Returns:
            A coroutine resolving when the history is cleared.
        """
        return self._rpc_request(account_id, {'type': 'removeApplication'})

    async def trade(self, account_id: str, trade) -> 'asyncio.Future[MetatraderTradeResponse]':
        """Execute a trade on a connected MetaTrader account
        (see https://metaapi.cloud/docs/client/websocket/api/trade/).

        Args:
            account_id: Id of the MetaTrader account to execute trade for.
            trade: Trade to execute (see docs for possible trade types).

        Returns:
            A coroutine resolving with trade result.

        Raises:
            TradeException: On trade error, check error properties for error code details.
        """
        response = await self._rpc_request(account_id, {'type': 'trade', 'trade': trade})
        if 'response' not in response:
            response['response'] = {}
        if 'stringCode' not in response['response']:
            response['response']['stringCode'] = response['response']['description']
        if 'numericCode' not in response['response']:
            response['response']['numericCode'] = response['response']['error']
        if response['response']['stringCode'] in ['ERR_NO_ERROR', 'TRADE_RETCODE_PLACED', 'TRADE_RETCODE_DONE',
                                                  'TRADE_RETCODE_DONE_PARTIAL', 'TRADE_RETCODE_NO_CHANGES']:
            return response['response']
        else:
            raise TradeException(response['response']['message'], response['response']['numericCode'],
                                 response['response']['stringCode'])

    def subscribe(self, account_id: str, instance_index: int = None):
        """Subscribes to the Metatrader terminal events
        (see https://metaapi.cloud/docs/client/websocket/api/subscribe/).

        Args:
            account_id: Id of the MetaTrader account to subscribe to.
            instance_index: Instance index.

        Returns:
            A coroutine which resolves when subscription started.
        """
        packet = {'type': 'subscribe'}
        if instance_index is not None:
            packet['instanceIndex'] = instance_index
        return self._rpc_request(account_id, packet)

    def reconnect(self, account_id: str) -> Coroutine:
        """Reconnects to the Metatrader terminal (see https://metaapi.cloud/docs/client/websocket/api/reconnect/).

        Args:
            account_id: Id of the MetaTrader account to reconnect.

        Returns:
            A coroutine which resolves when reconnection started
        """
        return self._rpc_request(account_id, {'type': 'reconnect'})

    def synchronize(self, account_id: str, instance_index: int, synchronization_id: str,
                    starting_history_order_time: datetime, starting_deal_time: datetime) -> Coroutine:
        """Requests the terminal to start synchronization process.
        (see https://metaapi.cloud/docs/client/websocket/synchronizing/synchronize/).

        Args:
            account_id: Id of the MetaTrader account to synchronize.
            instance_index: Instance index.
            synchronization_id: Synchronization request id.
            starting_history_order_time: From what date to start synchronizing history orders from. If not specified,
            the entire order history will be downloaded.
            starting_deal_time: From what date to start deal synchronization from. If not specified, then all
            history deals will be downloaded.

        Returns:
            A coroutine which resolves when synchronization is started.
        """
        return self._synchronizationThrottler.schedule_synchronize(account_id, {
            'requestId': synchronization_id, 'type': 'synchronize',
            'startingHistoryOrderTime': format_date(starting_history_order_time),
            'startingDealTime': format_date(starting_deal_time),
            'instanceIndex': instance_index})

    def wait_synchronized(self, account_id: str, instance_index: int, application_pattern: str,
                          timeout_in_seconds: float):
        """Waits for server-side terminal state synchronization to complete.
        (see https://metaapi.cloud/docs/client/websocket/synchronizing/waitSynchronized/).

        Args:
            account_id: Id of the MetaTrader account to synchronize.
            instance_index: Instance index.
            application_pattern: MetaApi application regular expression pattern, default is .*
            timeout_in_seconds: Timeout in seconds, default is 300 seconds.
        """
        return self._rpc_request(account_id, {'type': 'waitSynchronized', 'applicationPattern': application_pattern,
                                              'timeoutInSeconds': timeout_in_seconds, 'instanceIndex': instance_index},
                                 timeout_in_seconds + 1)

    def subscribe_to_market_data(self, account_id: str, instance_index: int, symbol: str) -> Coroutine:
        """Subscribes on market data of specified symbol
        (see https://metaapi.cloud/docs/client/websocket/marketDataStreaming/subscribeToMarketData/).

        Args:
            account_id: Id of the MetaTrader account.
            instance_index: Instance index.
            symbol: Symbol (e.g. currency pair or an index).

        Returns:
            A coroutine which resolves when subscription request was processed.
        """
        packet = {'type': 'subscribeToMarketData', 'symbol': symbol}
        if instance_index is not None:
            packet['instanceIndex'] = instance_index
        return self._rpc_request(account_id, packet)

    def unsubscribe_from_market_data(self, account_id: str, instance_index: int, symbol: str) -> Coroutine:
        """Unsubscribes from market data of specified symbol
        (see https://metaapi.cloud/docs/client/websocket/marketDataStreaming/unsubscribeFromMarketData/).

        Args:
            account_id: Id of the MetaTrader account.
            instance_index: Instance index.
            symbol: Symbol (e.g. currency pair or an index).

        Returns:
            A coroutine which resolves when unsubscription request was processed.
        """
        packet = {'type': 'unsubscribeFromMarketData', 'symbol': symbol}
        if instance_index is not None:
            packet['instanceIndex'] = instance_index
        return self._rpc_request(account_id, packet)

    async def get_symbol_specification(self, account_id: str, symbol: str) -> \
            'asyncio.Future[MetatraderSymbolSpecification]':
        """Retrieves specification for a symbol
        (see https://metaapi.cloud/docs/client/websocket/api/retrieveMarketData/getSymbolSpecification/).

        Args:
            account_id: Id of the MetaTrader account to retrieve symbol specification for.
            symbol: Symbol to retrieve specification for.

        Returns:
            A coroutine which resolves when specification is retrieved.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getSymbolSpecification',
                                                        'symbol': symbol})
        return response['specification']

    async def get_symbol_price(self, account_id: str, symbol: str) -> 'asyncio.Future[MetatraderSymbolPrice]':
        """Retrieves price for a symbol
        (see https://metaapi.cloud/docs/client/websocket/api/retrieveMarketData/getSymbolPrice/).

        Args:
            account_id: Id of the MetaTrader account to retrieve symbol price for.
            symbol: Symbol to retrieve price for.

        Returns:
            A coroutine which resolves when price is retrieved.
        """
        response = await self._rpc_request(account_id, {'application': 'RPC', 'type': 'getSymbolPrice',
                                                        'symbol': symbol})
        return response['price']

    def save_uptime(self, account_id: str, uptime: Dict):
        """Sends client uptime stats to the server.

        Args:
            account_id: Id of the MetaTrader account to retrieve symbol price for.
            uptime: Uptime statistics to send to the server.

        Returns:
            A coroutine which resolves when uptime statistics is submitted.
        """
        return self._rpc_request(account_id, {'type': 'saveUptime', 'uptime': uptime})

    def unsubscribe(self, account_id: str):
        """Unsubscribe from account (see https://metaapi.cloud/docs/client/websocket/api/synchronizing/unsubscribe).

        Args:
            account_id: Id of the MetaTrader account to retrieve symbol price for.

        Returns:
            A coroutine which resolves when socket is unsubscribed."""
        return self._rpc_request(account_id, {'type': 'unsubscribe'})

    def add_synchronization_listener(self, account_id: str, listener):
        """Adds synchronization listener for specific account.

        Args:
            account_id: Account id.
            listener: Synchronization listener to add.
        """
        if account_id in self._synchronizationListeners:
            listeners = self._synchronizationListeners[account_id]
        else:
            listeners = []
            self._synchronizationListeners[account_id] = listeners
        listeners.append(listener)

    def remove_synchronization_listener(self, account_id: str, listener: SynchronizationListener):
        """Removes synchronization listener for specific account.

        Args:
            account_id: Account id.
            listener: Synchronization listener to remove.
        """
        listeners = self._synchronizationListeners[account_id]

        if not listeners:
            listeners = []
        elif listeners.__contains__(listener):
            listeners.remove(listener)
        self._synchronizationListeners[account_id] = listeners

    def add_latency_listener(self, listener: LatencyListener):
        """Adds latency listener.

        Args:
            listener: Latency listener to add."""
        self._latencyListeners.append(listener)

    def remove_latency_listener(self, listener: LatencyListener):
        """Removes latency listener.

        Args:
            listener: Latency listener to remove."""
        self._latencyListeners = list(filter(lambda l: l != listener, self._latencyListeners))

    def add_reconnect_listener(self, listener: ReconnectListener):
        """Adds reconnect listener.

        Args:
            listener: Reconnect listener to add.
        """

        self._reconnectListeners.append(listener)

    def remove_reconnect_listener(self, listener: ReconnectListener):
        """Removes reconnect listener.

        Args:
            listener: Listener to remove.
        """

        if self._reconnectListeners.__contains__(listener):
            self._reconnectListeners.remove(listener)

    def remove_all_listeners(self):
        """Removes all listeners. Intended for use in unit tests."""

        self._synchronizationListeners = {}
        self._reconnectListeners = []

    async def _reconnect(self):
        reconnected = False
        while self._connected and not reconnected:
            try:
                await self._socket.disconnect()
                url = f'{self._url}?auth-token={self._token}'
                self._connectResult = asyncio.Future()
                self._resolved = False
                self._sessionId = random_id()
                await asyncio.wait_for(self._socket.connect(url, socketio_path='ws',
                                                            headers={'Client-id': '{:01.10f}'.format(random())}),
                                       timeout=self._connect_timeout)
                await self._connectResult
                reconnected = True
                await self._fire_reconnected()
                await self._socket.wait()
            except Exception:
                self._connectResult.cancel()
                self._connectResult = None
                await asyncio.sleep(1)

    async def _rpc_request(self, account_id: str, request: dict, timeout_in_seconds: float = None) -> Coroutine:
        if not self._connected:
            await self.connect()
        start_time = datetime.now()
        while not self._resolved and (start_time + timedelta(seconds=self._connect_timeout) > datetime.now()):
            await asyncio.sleep(1)
        if not self._resolved:
            raise TimeoutException(f"MetaApi websocket client request of account {account_id} timed because socket "
                                   f"client failed to connect to the server.")
        if request['type'] == 'subscribe':
            request['sessionId'] = self._sessionId
        if request['type'] in ['trade', 'subscribe']:
            return await self._make_request(account_id, request, timeout_in_seconds)
        retry_counter = 0
        while True:
            try:
                return await self._make_request(account_id, request, timeout_in_seconds)
            except Exception as err:
                if err.__class__.__name__ in ['NotSynchronizedException', 'TimeoutException',
                                              'NotAuthenticatedException', 'InternalException'] and retry_counter < \
                        self._retries:
                    await asyncio.sleep(min(pow(2, retry_counter) * self._minRetryDelayInSeconds,
                                            self._maxRetryDelayInSeconds))
                    retry_counter += 1
                else:
                    raise err

    async def _make_request(self, account_id: str, request: dict, timeout_in_seconds: float = None):
        if 'requestId' in request:
            request_id = request['requestId']
        else:
            request_id = random_id()
            request['requestId'] = request_id
        request['timestamps'] = {'clientProcessingStarted': format_date(datetime.now())}
        self._requestResolves[request_id] = asyncio.Future()
        self._requestResolves[request_id].type = request['type']
        request['accountId'] = account_id
        request['application'] = request['application'] if 'application' in request else self._application
        await self._socket.emit('request', request)
        try:
            resolve = await asyncio.wait_for(self._requestResolves[request_id], timeout=timeout_in_seconds or
                                             self._request_timeout)
        except asyncio.TimeoutError:
            raise TimeoutException(f"MetaApi websocket client request {request['requestId']} of type "
                                   f"{request['type']} timed out. Please make sure your account is connected "
                                   f"to broker before retrying your request.")
        return resolve

    def _convert_error(self, data) -> Exception:
        if data['error'] == 'ValidationError':
            return ValidationException(data['message'], data['details'])
        elif data['error'] == 'NotFoundError':
            return NotFoundException(data['message'])
        elif data['error'] == 'NotSynchronizedError':
            return NotSynchronizedException(data['message'])
        elif data['error'] == 'TimeoutError':
            return TimeoutException(data['message'])
        elif data['error'] == 'NotAuthenticatedError':
            return NotConnectedException(data['message'])
        elif data['error'] == 'TradeError':
            return TradeException(data['message'], data['numericCode'], data['stringCode'])
        elif data['error'] == 'UnauthorizedError':
            self.close()
            return UnauthorizedException(data['message'])
        else:
            return InternalException(data['message'])

    def _convert_iso_time_to_date(self, packet):
        if not isinstance(packet, str):
            for field in packet:
                value = packet[field]
                if isinstance(value, str) and re.search('time|Time', field) and not \
                        re.search('brokerTime|BrokerTime', field):
                    packet[field] = date(value)
                if isinstance(value, list):
                    for item in value:
                        self._convert_iso_time_to_date(item)
                if isinstance(value, dict):
                    self._convert_iso_time_to_date(value)
            if packet and 'timestamps' in packet:
                for field in packet['timestamps']:
                    packet['timestamps'][field] = date(packet['timestamps'][field])
            if packet and 'type' in packet and packet['type'] == 'prices':
                if 'prices' in packet:
                    for price in packet['prices']:
                        if 'timestamps' in price:
                            for field in price['timestamps']:
                                if isinstance(price['timestamps'][field], str):
                                    price['timestamps'][field] = date(price['timestamps'][field])

    async def _process_synchronization_packet(self, packet):
        try:
            packets = self._packetOrderer.restore_order(packet)
            for data in packets:
                if 'synchronizationId' in data:
                    self._synchronizationThrottler.update_synchronization_id(data['synchronizationId'])
                instance_id = data['accountId'] + ':' + str(data['instanceIndex'] if 'instanceIndex' in data else 0)
                instance_index = data['instanceIndex'] if 'instanceIndex' in data else 0
                if data['type'] == 'authenticated':
                    if 'sessionId' not in data or data['sessionId'] == self._sessionId:
                        if 'host' in data:
                            self._connectedHosts[instance_id] = data['host']

                        on_connected_tasks: List[asyncio.Task] = []

                        async def run_on_connected(listener):
                            try:
                                await listener.on_connected(instance_index, data['replicas'])
                            except Exception as err:
                                print(f'{data["accountId"]}: Failed to notify listener about connected event', err)

                        if data['accountId'] in self._synchronizationListeners:
                            for listener in self._synchronizationListeners[data['accountId']]:
                                on_connected_tasks.append(asyncio.create_task(run_on_connected(listener)))
                        if len(on_connected_tasks) > 0:
                            await asyncio.wait(on_connected_tasks)
                elif data['type'] == 'disconnected':
                    if instance_id in self._connectedHosts and self._connectedHosts[instance_id] == data['host']:
                        on_disconnected_tasks: List[asyncio.Task] = []

                        async def run_on_disconnected(listener):
                            try:
                                await listener.on_disconnected(instance_index)
                            except Exception as err:
                                print(f'{data["accountId"]}: Failed to notify listener about disconnected event', err)

                        if data['accountId'] in self._synchronizationListeners:
                            for listener in self._synchronizationListeners[data['accountId']]:
                                on_disconnected_tasks.append(asyncio.create_task(run_on_disconnected(listener)))
                        if len(on_disconnected_tasks) > 0:
                            await asyncio.wait(on_disconnected_tasks)
                        del self._connectedHosts[instance_id]
                elif data['type'] == 'synchronizationStarted':
                    on_sync_started_tasks: List[asyncio.Task] = []

                    async def run_on_sync_started(listener):
                        try:
                            await listener.on_synchronization_started(instance_index)
                        except Exception as err:
                            print(f'{data["accountId"]}: Failed to notify listener about synchronization started ' +
                                  'event', err)

                    if data['accountId'] in self._synchronizationListeners:
                        for listener in self._synchronizationListeners[data['accountId']]:
                            on_sync_started_tasks.append(asyncio.create_task(run_on_sync_started(listener)))
                    if len(on_sync_started_tasks) > 0:
                        await asyncio.wait(on_sync_started_tasks)
                elif data['type'] == 'accountInformation':
                    if data['accountInformation'] and (data['accountId'] in self._synchronizationListeners):
                        on_account_information_updated_tasks: List[asyncio.Task] = []

                        async def run_on_account_info(listener):
                            try:
                                await listener.on_account_information_updated(instance_index,
                                                                              data['accountInformation'])
                            except Exception as err:
                                print(f'{data["accountId"]}: Failed to notify listener about ' +
                                      'accountInformation event', err)

                        for listener in self._synchronizationListeners[data['accountId']]:
                            on_account_information_updated_tasks.append(asyncio
                                                                        .create_task(run_on_account_info(listener)))
                        if len(on_account_information_updated_tasks) > 0:
                            await asyncio.wait(on_account_information_updated_tasks)
                elif data['type'] == 'deals':
                    if 'deals' in data:
                        for deal in data['deals']:
                            on_deal_added_tasks: List[asyncio.Task] = []

                            async def run_on_deal_added(listener):
                                try:
                                    await listener.on_deal_added(instance_index, deal)
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about deals event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_deal_added_tasks.append(asyncio.create_task(run_on_deal_added(listener)))
                            if len(on_deal_added_tasks) > 0:
                                await asyncio.wait(on_deal_added_tasks)
                elif data['type'] == 'orders':
                    on_order_updated_tasks: List[asyncio.Task] = []

                    async def run_on_order_updated(listener):
                        try:
                            if 'orders' in data:
                                await listener.on_orders_replaced(instance_index, data['orders'])
                        except Exception as err:
                            print(f'{data["accountId"]}: Failed to notify listener about orders event', err)

                    if data['accountId'] in self._synchronizationListeners:
                        for listener in self._synchronizationListeners[data['accountId']]:
                            on_order_updated_tasks.append(asyncio.create_task(run_on_order_updated(listener)))
                    if len(on_order_updated_tasks) > 0:
                        await asyncio.wait(on_order_updated_tasks)
                elif data['type'] == 'historyOrders':
                    if 'historyOrders' in data:
                        for historyOrder in data['historyOrders']:
                            on_history_order_added_tasks: List[asyncio.Task] = []

                            async def run_on_order_added(listener):
                                try:
                                    await listener.on_history_order_added(instance_index, historyOrder)
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about historyOrders ' +
                                          'event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_history_order_added_tasks.append(asyncio
                                                                        .create_task(run_on_order_added(listener)))
                            if len(on_history_order_added_tasks) > 0:
                                await asyncio.wait(on_history_order_added_tasks)
                elif data['type'] == 'positions':
                    on_position_updated_tasks: List[asyncio.Task] = []

                    async def run_on_position_updated(listener):
                        try:
                            if 'positions' in data:
                                await listener.on_positions_replaced(instance_index, data['positions'])
                        except Exception as err:
                            print(f'{data["accountId"]}: Failed to notify listener about positions event', err)

                    if data['accountId'] in self._synchronizationListeners:
                        for listener in self._synchronizationListeners[data['accountId']]:
                            on_position_updated_tasks.append(asyncio
                                                             .create_task(run_on_position_updated(listener)))
                    if len(on_position_updated_tasks) > 0:
                        await asyncio.wait(on_position_updated_tasks)
                elif data['type'] == 'update':
                    if 'accountInformation' in data and (data['accountId'] in self._synchronizationListeners):
                        on_account_information_updated_tasks: List[asyncio.Task] = []

                        async def run_on_account_information_updated(listener):
                            try:
                                await listener.on_account_information_updated(instance_index,
                                                                              data['accountInformation'])
                            except Exception as err:
                                print(f'{data["accountId"]}: Failed to notify listener about update event', err)

                        for listener in self._synchronizationListeners[data['accountId']]:
                            on_account_information_updated_tasks.append(
                                asyncio.create_task(run_on_account_information_updated(listener)))
                        if len(on_account_information_updated_tasks) > 0:
                            await asyncio.wait(on_account_information_updated_tasks)
                    if 'updatedPositions' in data:
                        for position in data['updatedPositions']:
                            on_position_updated_tasks: List[asyncio.Task] = []

                            async def run_on_position_updated(listener):
                                try:
                                    await listener.on_position_updated(instance_index, position)
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about update event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_position_updated_tasks.append(
                                        asyncio.create_task(run_on_position_updated(listener)))
                            if len(on_position_updated_tasks) > 0:
                                await asyncio.wait(on_position_updated_tasks)
                    if 'removedPositionIds' in data:
                        for positionId in data['removedPositionIds']:
                            on_position_removed_tasks: List[asyncio.Task] = []

                            async def run_on_position_removed(listener):
                                try:
                                    await listener.on_position_removed(instance_index, positionId)
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about update event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_position_removed_tasks.append(
                                        asyncio.create_task(run_on_position_removed(listener)))
                            if len(on_position_removed_tasks) > 0:
                                await asyncio.wait(on_position_removed_tasks)
                    if 'updatedOrders' in data:
                        for order in data['updatedOrders']:
                            on_order_updated_tasks: List[asyncio.Task] = []

                            async def run_on_order_updated(listener):
                                try:
                                    await listener.on_order_updated(instance_index, order)
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about update event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_order_updated_tasks.append(
                                        asyncio.create_task(run_on_order_updated(listener)))
                            if len(on_order_updated_tasks) > 0:
                                await asyncio.wait(on_order_updated_tasks)
                    if 'completedOrderIds' in data:
                        for orderId in data['completedOrderIds']:
                            on_order_completed_tasks: List[asyncio.Task] = []

                            async def run_on_order_completed(listener):
                                try:
                                    await listener.on_order_completed(instance_index, orderId)
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about update event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_order_completed_tasks.append(
                                        asyncio.create_task(run_on_order_completed(listener)))
                            if len(on_order_completed_tasks) > 0:
                                await asyncio.wait(on_order_completed_tasks)
                    if 'historyOrders' in data:
                        for historyOrder in data['historyOrders']:
                            on_history_order_added_tasks: List[asyncio.Task] = []

                            async def run_on_history_order_added(listener):
                                try:
                                    await listener.on_history_order_added(instance_index, historyOrder)
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about update event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_history_order_added_tasks.append(
                                        asyncio.create_task(run_on_history_order_added(listener)))
                            if len(on_history_order_added_tasks) > 0:
                                await asyncio.wait(on_history_order_added_tasks)
                    if 'deals' in data:
                        for deal in data['deals']:
                            on_deal_added_tasks: List[asyncio.Task] = []

                            async def run_on_deal_added(listener):
                                try:
                                    await listener.on_deal_added(instance_index, deal)
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about update event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_deal_added_tasks.append(
                                        asyncio.create_task(run_on_deal_added(listener)))
                            if len(on_deal_added_tasks) > 0:
                                await asyncio.wait(on_deal_added_tasks)
                    if 'timestamps' in data:
                        data['timestamps']['clientProcessingFinished'] = datetime.now()
                        on_update_tasks: List[asyncio.Task] = []

                        async def run_on_update(listener):
                            try:
                                await listener.on_update(data['accountId'], data['timestamps'])
                            except Exception as err:
                                print(f'{data["accountId"]}: Failed to notify latency listener about update event', err)

                        for listener in self._latencyListeners:
                            on_update_tasks.append(asyncio.create_task(run_on_update(listener)))

                        if len(on_update_tasks) > 0:
                            await asyncio.wait(on_update_tasks)
                elif data['type'] == 'dealSynchronizationFinished':
                    if data['accountId'] in self._synchronizationListeners:
                        on_deal_synchronization_finished_tasks: List[asyncio.Task] = []

                        async def run_on_deal_synchronization_finished(listener):
                            self._synchronizationThrottler.remove_synchronization_id(data['synchronizationId'])
                            try:
                                await listener.on_deal_synchronization_finished(instance_index,
                                                                                data['synchronizationId'])
                            except Exception as err:
                                print(f'{data["accountId"]}: Failed to notify listener about ' +
                                      'dealSynchronizationFinished event', err)

                        for listener in self._synchronizationListeners[data['accountId']]:
                            on_deal_synchronization_finished_tasks.append(
                                        asyncio.create_task(run_on_deal_synchronization_finished(listener)))
                        if len(on_deal_synchronization_finished_tasks) > 0:
                            await asyncio.wait(on_deal_synchronization_finished_tasks)
                elif data['type'] == 'orderSynchronizationFinished':
                    if data['accountId'] in self._synchronizationListeners:
                        on_order_synchronization_finished_tasks: List[asyncio.Task] = []

                        async def run_on_order_synchronization_finished(listener):
                            try:
                                await listener.on_order_synchronization_finished(instance_index,
                                                                                 data['synchronizationId'])
                            except Exception as err:
                                print(f'{data["accountId"]}: Failed to notify listener about ' +
                                      'orderSynchronizationFinished event', err)

                        for listener in self._synchronizationListeners[data['accountId']]:
                            on_order_synchronization_finished_tasks.append(
                                asyncio.create_task(run_on_order_synchronization_finished(listener)))
                        if len(on_order_synchronization_finished_tasks) > 0:
                            await asyncio.wait(on_order_synchronization_finished_tasks)
                elif data['type'] == 'status':
                    if instance_id not in self._connectedHosts:
                        if instance_id not in self._resubscriptionTriggerTimes:
                            self._resubscriptionTriggerTimes[instance_id] = datetime.now()
                        elif self._resubscriptionTriggerTimes[instance_id].timestamp() + 2 * 60 < \
                                datetime.now().timestamp():
                            del self._resubscriptionTriggerTimes[instance_id]
                            print(f'[{datetime.now().isoformat()}] it seems like we are not connected to a ' +
                                  'running API server yet, retrying subscription for account ' + instance_id)

                            async def run_subscribe():
                                try:
                                    await self.subscribe(data['accountId'], data['instanceIndex'])
                                except Exception as error:
                                    print(f'[{datetime.now().isoformat()}] MetaApi websocket client failed to ' +
                                          'receive subscribe response for account id ' + instance_id, error)
                            asyncio.create_task(run_subscribe())
                    elif instance_id in self._connectedHosts and self._connectedHosts[instance_id] == data['host']:
                        if instance_id in self._resubscriptionTriggerTimes:
                            del self._resubscriptionTriggerTimes[instance_id]
                        on_broker_connection_status_changed_tasks: List[asyncio.Task] = []

                        async def run_on_broker_connection_status_changed(listener):
                            try:
                                await listener.on_broker_connection_status_changed(instance_index,
                                                                                   bool(data['connected']))
                            except Exception as err:
                                print(f'{data["accountId"]}: Failed to notify listener about ' +
                                      'brokerConnectionStatusChanged event', err)

                        for listener in self._synchronizationListeners[data['accountId']]:
                            on_broker_connection_status_changed_tasks.append(
                                asyncio.create_task(run_on_broker_connection_status_changed(listener)))
                        if len(on_broker_connection_status_changed_tasks) > 0:
                            await asyncio.wait(on_broker_connection_status_changed_tasks)
                        if 'healthStatus' in data:
                            on_health_status_tasks: List[asyncio.Task] = []

                            async def run_on_health_status(listener):
                                try:
                                    await listener.on_health_status(instance_index, data['healthStatus'])
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about server-side ' +
                                          'healthStatus event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_health_status_tasks.append(
                                        asyncio.create_task(run_on_health_status(listener)))
                                if len(on_health_status_tasks) > 0:
                                    await asyncio.wait(on_health_status_tasks)
                elif data['type'] == 'specifications':
                    if 'specifications' in data:
                        for specification in data['specifications']:
                            on_symbol_specification_updated_tasks: List[asyncio.Task] = []

                            async def run_on_symbol_specification_updated(listener):
                                try:
                                    await listener.on_symbol_specification_updated(instance_index, specification)
                                except Exception as err:
                                    print(f'{data["accountId"]}: Failed to notify listener about specifications ' +
                                          'event', err)

                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_symbol_specification_updated_tasks.append(
                                        asyncio.create_task(run_on_symbol_specification_updated(listener)))
                                if len(on_symbol_specification_updated_tasks) > 0:
                                    await asyncio.wait(on_symbol_specification_updated_tasks)
                elif data['type'] == 'prices':
                    if 'prices' in data:
                        prices = data['prices']
                        on_symbol_prices_updated_tasks: List[asyncio.Task] = []

                        async def run_on_symbol_prices_updated(listener):
                            try:
                                equity = data['equity'] if 'equity' in data else None
                                margin = data['margin'] if 'margin' in data else None
                                free_margin = data['freeMargin'] if 'freeMargin' in data else None
                                margin_level = data['marginLevel'] if 'marginLevel' in data else None
                                await listener.on_symbol_prices_updated(instance_index, prices, equity, margin,
                                                                        free_margin, margin_level)
                            except Exception as err:
                                print(f'{data["accountId"]}: Failed to notify listener about prices event', err)

                        if data['accountId'] in self._synchronizationListeners:
                            for listener in self._synchronizationListeners[data['accountId']]:
                                on_symbol_prices_updated_tasks.append(
                                    asyncio.create_task(run_on_symbol_prices_updated(listener)))
                            if len(on_symbol_prices_updated_tasks) > 0:
                                await asyncio.wait(on_symbol_prices_updated_tasks)
                        for price in data['prices']:
                            on_symbol_price_updated_tasks: List[asyncio.Task] = []

                            async def run_on_symbol_price_updated(listener):
                                try:
                                    await listener.on_symbol_price_updated(instance_index, price)
                                except Exception as err:
                                    print('Failed to notify listener about price event', err)
                            if data['accountId'] in self._synchronizationListeners:
                                for listener in self._synchronizationListeners[data['accountId']]:
                                    on_symbol_price_updated_tasks.append(
                                        asyncio.create_task(run_on_symbol_price_updated(listener)))
                                if len(on_symbol_price_updated_tasks) > 0:
                                    await asyncio.wait(on_symbol_price_updated_tasks)

                        for price in data['prices']:
                            if 'timestamps' in price:
                                price['timestamps']['clientProcessingFinished'] = datetime.now()
                                on_symbol_price_tasks: List[asyncio.Task] = []

                                async def run_on_symbol_price(listener):
                                    try:
                                        await listener.on_symbol_price(data['accountId'], price['symbol'],
                                                                       price['timestamps'])
                                    except Exception as err:
                                        print(f'{data["accountId"]}: Failed to notify latency listener about ' +
                                              'update event', err)
                                for listener in self._latencyListeners:
                                    on_symbol_price_tasks.append(
                                        asyncio.create_task(run_on_symbol_price(listener)))
                                if len(on_symbol_price_tasks) > 0:
                                    await asyncio.wait(on_symbol_price_tasks)
        except Exception as err:
            print('Failed to process incoming synchronization packet', err)

    async def _fire_reconnected(self):
        for listener in self._reconnectListeners:
            try:
                await listener.on_reconnected()
            except Exception as err:
                print(f'[{datetime.now().isoformat()}] Failed to notify reconnect listener', err)



from .metatraderAccountApi import MetatraderAccountApi
from .metatraderAccount import MetatraderAccount
from ..clients.errorHandler import NotFoundException
from ..clients.metaApi.metaApiWebsocket_client import MetaApiWebsocketClient
from ..clients.metaApi.metatraderAccount_client import MetatraderAccountClient, NewMetatraderAccountDto
from .metaApiConnection import MetaApiConnection
from ..clients.metaApi.reconnectListener import ReconnectListener
from .connectionRegistry import ConnectionRegistry
from .memoryHistoryStorage import MemoryHistoryStorage
from .historyStorage import HistoryStorage
from mock import AsyncMock, MagicMock, patch
from .metatraderAccountModel import MetatraderAccountModel
from httpx import Response
from datetime import datetime
from .models import date
import pytest


class MockClient(MetatraderAccountClient):
    def get_accounts(self, provisioning_profile_id: str = None) -> Response:
        pass

    def get_account(self, id: str) -> Response:
        pass

    def create_account(self, account: NewMetatraderAccountDto) -> Response:
        pass

    def delete_account(self, id: str) -> Response:
        pass

    def deploy_account(self, id: str) -> Response:
        pass

    def undeploy_account(self, id: str) -> Response:
        pass

    def redeploy_account(self, id: str) -> Response:
        pass


class MockWebsocketClient(MetaApiWebsocketClient):
    def add_synchronization_listener(self, account_id: str, listener):
        pass

    def add_reconnect_listener(self, listener: ReconnectListener):
        pass

    def subscribe(self, account_id: str):
        pass


class MockStorage(MemoryHistoryStorage):
    async def last_history_order_time(self) -> datetime:
        return date('2020-01-01T00:00:00.000Z')

    async def last_deal_time(self) -> datetime:
        return date('2020-01-02T00:00:00.000Z')


class MockRegistry(ConnectionRegistry):
    async def connect(self, account: MetatraderAccountModel, history_storage: HistoryStorage,
                      history_start_time: datetime = None):
        pass

    def remove(self, account_id: str):
        pass


client: MockClient = None
websocket_client: MockWebsocketClient = None
registry: MockRegistry = None
api: MetatraderAccountApi = None


@pytest.fixture(autouse=True)
async def run_around_tests():
    global client
    client = MockClient(MagicMock(), MagicMock())
    global websocket_client
    websocket_client = MockWebsocketClient('token')
    global registry
    registry = MockRegistry(websocket_client)
    global api
    registry.connect = AsyncMock()
    registry.remove = MagicMock()
    api = MetatraderAccountApi(client, websocket_client, registry)
    yield


class TestMetatraderAccountApi:
    @pytest.mark.asyncio
    async def test_retrieve_mt_accounts(self):
        """Should retrieve MT accounts."""
        client.get_accounts = AsyncMock(return_value=[{'_id': 'id'}])
        accounts = await api.get_accounts({'provisioningProfileId': 'profileId'})
        assert list(map(lambda a: a.id, accounts)) == ['id']
        for account in accounts:
            assert isinstance(account, MetatraderAccount)
        client.get_accounts.assert_called_with({'provisioningProfileId': 'profileId'})

    @pytest.mark.asyncio
    async def test_retrieve_mt_account_by_id(self):
        """Should retrieve MT account by id."""
        client.get_account = AsyncMock(return_value={
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud',
            'accessToken': '2RUnoH1ldGbnEneCoqRTgI4QO1XOmVzbH5EVoQsA'
        })
        account = await api.get_account('id')
        assert account.id == 'id'
        assert account.login == '50194988'
        assert account.name == 'mt5a'
        assert account.server == 'ICMarketsSC-Demo'
        assert account.provisioning_profile_id == 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076'
        assert account.magic == 123456
        assert account.application == 'MetaApi'
        assert account.connection_status == 'DISCONNECTED'
        assert account.state == 'DEPLOYED'
        assert account.type == 'cloud'
        assert account.access_token == '2RUnoH1ldGbnEneCoqRTgI4QO1XOmVzbH5EVoQsA'
        assert isinstance(account, MetatraderAccount)
        client.get_account.assert_called_with('id')

    @pytest.mark.asyncio
    async def test_retrieve_mt_account_by_token(self):
        """Should retrieve MT account by id."""
        client.get_account_by_token = AsyncMock(return_value={
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud',
            'accessToken': '2RUnoH1ldGbnEneCoqRTgI4QO1XOmVzbH5EVoQsA'
        })
        account = await api.get_account_by_token()
        assert account.id == 'id'
        assert account.login == '50194988'
        assert account.name == 'mt5a'
        assert account.server == 'ICMarketsSC-Demo'
        assert account.provisioning_profile_id == 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076'
        assert account.magic == 123456
        assert account.application == 'MetaApi'
        assert account.connection_status == 'DISCONNECTED'
        assert account.state == 'DEPLOYED'
        assert account.type == 'cloud'
        assert account.access_token == '2RUnoH1ldGbnEneCoqRTgI4QO1XOmVzbH5EVoQsA'
        assert isinstance(account, MetatraderAccount)
        client.get_account_by_token.assert_called_with()

    @pytest.mark.asyncio
    async def test_create_mt_account(self):
        """Should create MT account."""
        client.create_account = AsyncMock(return_value={'id': 'id'})
        client.get_account = AsyncMock(return_value={
          '_id': 'id',
          'login': '50194988',
          'name': 'mt5a',
          'server': 'ICMarketsSC-Demo',
          'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
          'magic': 123456,
          'application': 'MetaApi',
          'connectionStatus': 'DISCONNECTED',
          'state': 'DEPLOYED',
          'type': 'cloud',
          'accessToken': '2RUnoH1ldGbnEneCoqRTgI4QO1XOmVzbH5EVoQsA'
        })
        new_account_data = {
            'login': '50194988',
            'password': 'Test1234',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'type': 'cloud',
            'accessToken': 'NyV5no9TMffJyUts2FjI80wly0so3rVCz4xOqiDx'
        }
        account = await api.create_account(new_account_data)
        assert account.id == 'id'
        assert account.login == '50194988'
        assert account.name == 'mt5a'
        assert account.server == 'ICMarketsSC-Demo'
        assert account.provisioning_profile_id == 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076'
        assert account.magic == 123456
        assert account.application == 'MetaApi'
        assert account.connection_status == 'DISCONNECTED'
        assert account.state == 'DEPLOYED'
        assert account.type == 'cloud'
        assert account.access_token == '2RUnoH1ldGbnEneCoqRTgI4QO1XOmVzbH5EVoQsA'
        assert isinstance(account, MetatraderAccount)
        client.create_account.assert_called_with(new_account_data)
        client.get_account.assert_called_with('id')

    @pytest.mark.asyncio
    async def test_reload_mt_account(self):
        """Should reload MT account."""
        client.get_account = AsyncMock(side_effect=[{
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYING',
            'type': 'cloud'
          },
            {
                '_id': 'id',
                'login': '50194988',
                'name': 'mt5a',
                'server': 'ICMarketsSC-Demo',
                'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
                'magic': 123456,
                'application': 'MetaApi',
                'connectionStatus': 'CONNECTED',
                'state': 'DEPLOYED',
                'type': 'cloud'
            }])
        account = await api.get_account('id')
        await account.reload()
        assert account.connection_status == 'CONNECTED'
        assert account.state == 'DEPLOYED'
        client.get_account.assert_called_with('id')
        assert client.get_account.call_count == 2

    @pytest.mark.asyncio
    async def test_remove_mt_account(self):
        """Should remove MT account."""
        with patch('lib.metaApi.metatraderAccount.HistoryFileManager.delete_storage_from_disk',
                   new_callable=AsyncMock) as delete_mock:
            client.get_account = AsyncMock(side_effect=[{
                '_id': 'id',
                'login': '50194988',
                'name': 'mt5a',
                'server': 'ICMarketsSC-Demo',
                'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
                'magic': 123456,
                'application': 'MetaApi',
                'connectionStatus': 'CONNECTED',
                'state': 'DEPLOYED',
                'type': 'cloud'
              },
                {
                    '_id': 'id',
                    'login': '50194988',
                    'name': 'mt5a',
                    'server': 'ICMarketsSC-Demo',
                    'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
                    'magic': 123456,
                    'application': 'MetaApi',
                    'connectionStatus': 'CONNECTED',
                    'state': 'DELETING',
                    'type': 'cloud'
                }
            ])
            client.delete_account = AsyncMock()
            account = await api.get_account('id')
            await account.remove()
            delete_mock.assert_called()
            registry.remove.assert_called_with('id')
            assert account.state == 'DELETING'
            client.delete_account.assert_called_with('id')
            client.get_account.assert_called_with('id')
            assert client.get_account.call_count == 2

    @pytest.mark.asyncio
    async def test_deploy_mt_account(self):
        """Should deploy MT account."""
        client.get_account = AsyncMock(side_effect=[{
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'UNDEPLOYED',
            'type': 'cloud'
          }, {
                '_id': 'id',
                'login': '50194988',
                'name': 'mt5a',
                'server': 'ICMarketsSC-Demo',
                'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
                'magic': 123456,
                'application': 'MetaApi',
                'connectionStatus': 'CONNECTED',
                'state': 'DEPLOYING',
                'type': 'cloud'
        }])
        client.deploy_account = AsyncMock()
        account = await api.get_account('id')
        await account.deploy()
        assert account.state == 'DEPLOYING'
        client.deploy_account.assert_called_with('id')
        client.get_account.assert_called_with('id')
        assert client.get_account.call_count == 2

    @pytest.mark.asyncio
    async def test_undeploy_mt_account(self):
        """Should undeploy MT account."""
        client.get_account = AsyncMock(side_effect=[{
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud'
          }, {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'CONNECTED',
            'state': 'UNDEPLOYING',
            'type': 'cloud'
        }])
        client.undeploy_account = AsyncMock()
        account = await api.get_account('id')
        await account.undeploy()
        registry.remove.assert_called_with('id')
        assert account.state == 'UNDEPLOYING'
        client.undeploy_account.assert_called_with('id')
        client.get_account.assert_called_with('id')
        assert client.get_account.call_count == 2

    @pytest.mark.asyncio
    async def test_redeploy_mt_account(self):
        """Should redeploy MT account."""
        client.get_account = AsyncMock(side_effect=[{
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud'
          }, {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'CONNECTED',
            'state': 'UNDEPLOYING',
            'type': 'cloud'
          }])
        client.redeploy_account = AsyncMock()
        account = await api.get_account('id')
        await account.redeploy()
        assert account.state == 'UNDEPLOYING'
        client.redeploy_account.assert_called_with('id')
        client.get_account.assert_called_with('id')
        assert client.get_account.call_count == 2

    @pytest.mark.asyncio
    async def test_wait_for_deployment(self):
        """Should wait for deployment."""
        deploying_account = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYING',
            'type': 'cloud'
        }
        client.get_account = AsyncMock(side_effect=[deploying_account, deploying_account,
                                                    {
                                                        '_id': 'id',
                                                        'login': '50194988',
                                                        'name': 'mt5a',
                                                        'server': 'ICMarketsSC-Demo',
                                                        'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
                                                        'magic': 123456,
                                                        'application': 'MetaApi',
                                                        'connectionStatus': 'CONNECTED',
                                                        'state': 'DEPLOYED',
                                                        'type': 'cloud'
                                                    }
                                                    ])
        account = await api.get_account('id')
        await account.wait_deployed(1, 50)
        assert account.state == 'DEPLOYED'
        client.get_account.assert_called_with('id')
        assert client.get_account.call_count == 3

    @pytest.mark.asyncio
    async def test_time_out_deployment(self):
        """Should time out waiting for deployment."""
        deploying_account = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYING',
            'type': 'cloud'
        }
        client.get_account = AsyncMock(return_value=deploying_account)
        account = await api.get_account('id')
        try:
            await account.wait_deployed(1, 50)
            raise Exception('TimeoutError is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
            assert account.state == 'DEPLOYING'
        client.get_account.assert_called_with('id')

    @pytest.mark.asyncio
    async def test_wait_for_undeployment(self):
        """Should wait for undeployment."""
        undeploying_account = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'UNDEPLOYING',
            'type': 'cloud'
        }
        client.get_account = AsyncMock(side_effect=[undeploying_account, undeploying_account,
                                                    {
                                                        '_id': 'id',
                                                        'login': '50194988',
                                                        'name': 'mt5a',
                                                        'server': 'ICMarketsSC-Demo',
                                                        'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
                                                        'magic': 123456,
                                                        'application': 'MetaApi',
                                                        'connectionStatus': 'CONNECTED',
                                                        'state': 'UNDEPLOYED',
                                                        'type': 'cloud'
                                                    }
                                                    ])
        account = await api.get_account('id')
        await account.wait_undeployed(1, 50)
        assert account.state == 'UNDEPLOYED'
        client.get_account.assert_called_with('id')
        assert client.get_account.call_count == 3

    @pytest.mark.asyncio
    async def test_time_out_undeployment(self):
        """Should wait for undeployment."""
        undeploying_account = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'UNDEPLOYING',
            'type': 'cloud'
        }
        client.get_account = AsyncMock(return_value=undeploying_account)
        account = await api.get_account('id')
        try:
            await account.wait_undeployed(1, 50)
            raise Exception('TimeoutException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
            assert account.state == 'UNDEPLOYING'
        client.get_account.assert_called_with('id')

    @pytest.mark.asyncio
    async def test_wait_until_removed(self):
        """Should wait until removed."""
        deleting_account = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DELETING',
            'type': 'cloud'
          }
        client.get_account = AsyncMock(side_effect=[deleting_account, deleting_account, NotFoundException('')])
        account = await api.get_account('id')
        await account.wait_removed(1, 50)
        client.get_account.assert_called_with('id')
        assert client.get_account.call_count == 3

    @pytest.mark.asyncio
    async def test_time_out_waiting_until_removed(self):
        """Should wait until removed."""
        deleting_account = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DELETING',
            'type': 'cloud'
        }
        client.get_account = AsyncMock(return_value=deleting_account)
        account = await api.get_account('id')
        try:
            await account.wait_removed(1, 50)
            raise Exception('TimeoutException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
        client.get_account.assert_called_with('id')

    @pytest.mark.asyncio
    async def test_wait_until_broker_connection(self):
        """Should wait util broker connection."""
        disconnected_account = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud'
        }
        client.get_account = AsyncMock(side_effect=[disconnected_account, disconnected_account,
                                                    {
                                                        '_id': 'id',
                                                        'login': '50194988',
                                                        'name': 'mt5a',
                                                        'server': 'ICMarketsSC-Demo',
                                                        'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
                                                        'magic': 123456,
                                                        'application': 'MetaApi',
                                                        'connectionStatus': 'CONNECTED',
                                                        'state': 'DEPLOYED',
                                                        'type': 'cloud'
                                                    }])
        account = await api.get_account('id')
        await account.wait_connected(1, 50)
        assert account.connection_status == 'CONNECTED'
        client.get_account.assert_called_with('id')
        assert client.get_account.call_count == 3

    @pytest.mark.asyncio
    async def test_time_out_waiting_for_broker_connection(self):
        """Should time out waiting for broker connection."""
        disconnected_account = {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'DISCONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud'
        }
        client.get_account = AsyncMock(return_value=disconnected_account)
        account = await api.get_account('id')
        try:
            await account.wait_connected(1, 50)
            raise Exception('TimeoutException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
            assert account.connection_status == 'DISCONNECTED'
        client.get_account.assert_called_with('id')

    @pytest.mark.asyncio
    async def test_connect_to_mt_terminal(self):
        """Should connect to an MT terminal."""
        websocket_client.add_synchronization_listener = MagicMock()
        websocket_client.subscribe = AsyncMock()
        MetaApiConnection.initialize = AsyncMock()
        client.get_account = AsyncMock(return_value={'_id': 'id'})
        account = await api.get_account('id')
        storage = MockStorage('accountId')
        await account.connect(storage)
        registry.connect.assert_called_with(account, storage, None)

    @pytest.mark.asyncio
    async def test_update_mt_account(self):
        """Should update MT account."""
        client.get_account = AsyncMock(side_effect=[{
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a',
            'server': 'ICMarketsSC-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'CONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud'
          }, {
            '_id': 'id',
            'login': '50194988',
            'name': 'mt5a__',
            'server': 'OtherMarkets-Demo',
            'provisioningProfileId': 'f9ce1f12-e720-4b9a-9477-c2d4cb25f076',
            'magic': 123456,
            'application': 'MetaApi',
            'connectionStatus': 'CONNECTED',
            'state': 'DEPLOYED',
            'type': 'cloud'
          }])
        client.update_account = AsyncMock()
        account = await api.get_account('id')
        await account.update({
          'name': 'mt5a__',
          'password': 'moreSecurePass',
          'server': 'OtherMarkets-Demo',
        })
        assert account.name == 'mt5a__'
        assert account.server == 'OtherMarkets-Demo'
        client.update_account.assert_called_with('id', {
          'name': 'mt5a__',
          'password': 'moreSecurePass',
          'server': 'OtherMarkets-Demo',
        })
        client.get_account.assert_called_with('id')
        assert client.get_account.call_count == 2



class NotConnectedException(Exception):
    """Exception which indicates that MetaTrader terminal did not start yet. You need to wait until account is
    connected and retry.
    """

    def __init__(self, message: str):
        """Inits the exception.

        Args:
            message: Exception message.
        """
        super().__init__(message)



from .httpClient import HttpClient
import re
import pytest
import respx
import asyncio
from httpx import Response, ConnectTimeout, Request
from .errorHandler import ApiException
httpClient = None
test_url = 'http://example.com'


@pytest.fixture(autouse=True)
async def run_around_tests():
    global httpClient
    httpClient = HttpClient(retry_opts={'minDelayInSeconds': 0.05, 'maxDelayInSeconds': 0.2})
    yield


class TestHttpClient:
    @pytest.mark.asyncio
    async def test_load(self):
        """Should load HTML page from example.com"""
        opts = {
            'url': test_url
        }
        response = await httpClient.request(opts)
        text = response.text
        assert re.search('doctype html', text)

    @pytest.mark.asyncio
    async def test_not_found(self):
        """Should return NotFound exception if server returns 404"""
        opts = {
            'url': f'{test_url}/not-found'
        }
        try:
            await httpClient.request(opts)
            raise Exception('NotFoundException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'NotFoundException'

    @pytest.mark.asyncio
    async def test_timeout(self):
        """Should return ConnectTimeout exception if request is timed out"""
        httpClient = HttpClient(0.001, {'retries': 2, 'minDelayInSeconds': 0.05, 'maxDelayInSeconds': 0.2})
        opts = {
            'url': test_url
        }
        try:
            await httpClient.request(opts)
            raise Exception('ConnectTimeout is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ConnectTimeout'

    @respx.mock
    @pytest.mark.asyncio
    async def test_validation_exception(self):
        """Should return a validation exception"""
        error = {
            'id': 1,
            'error': 'error',
            'message': 'test message',
        }
        respx.post(test_url).mock(return_value=Response(400, json=error))
        opts = {
            'method': 'POST',
            'url': test_url
        }
        try:
            await httpClient.request(opts)
            raise Exception('ValidationException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'
            assert err.__str__() == 'test message, check error.details for more information'

    @respx.mock
    @pytest.mark.asyncio
    async def test_validation_exception_details(self):
        """Should return a validation exception with details"""
        error = {
            'id': 1,
            'error': 'error',
            'message': 'test',
            'details': [{'parameter': 'password', 'value': 'wrong', 'message': 'Invalid value'}]
        }
        respx.post(test_url).mock(return_value=Response(400, json=error))
        opts = {
            'method': 'POST',
            'url': test_url
        }
        try:
            await httpClient.request(opts)
            raise Exception('ValidationException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'
            assert err.__str__() == 'test, check error.details for more information'
            assert err.details == error['details']

    @respx.mock
    @pytest.mark.asyncio
    async def test_retry_on_fail(self):
        """Should retry request on fail."""
        opts = {
            'method': 'POST',
            'url': test_url
        }
        respx.post(test_url).mock(side_effect=[ConnectTimeout('test', request=Request('GET', opts['url'])),
                                               ConnectTimeout('test', request=Request('GET', opts['url'])),
                                               Response(200, text='response')])
        response = await httpClient.request(opts)
        assert response.text == 'response'

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_on_retry_limit_exceeded(self):
        """Should return error if retry limit exceeded."""
        opts = {
            'method': 'POST',
            'url': test_url
        }
        respx.post(test_url).mock(side_effect=ConnectTimeout)
        try:
            await httpClient.request(opts)
            raise Exception('ConnectTimeout is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ConnectTimeout'

    @respx.mock
    @pytest.mark.asyncio
    async def test_not_retry_if_error_not_specified(self):
        """Should not retry if error not specified."""
        error = {
            'id': 1,
            'error': 'error',
            'message': 'test',
            'details': [{'parameter': 'password', 'value': 'wrong', 'message': 'Invalid value'}]
        }
        respx.post(test_url).mock(side_effect=[Response(400, json=error), Response(400, json=error), Response(204)])
        opts = {
            'method': 'POST',
            'url': test_url
        }
        try:
            await httpClient.request(opts)
            raise Exception('ValidationException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'
            assert err.__str__() == 'test, check error.details for more information'
            assert err.details == error['details']



def create_new_node_(key):
    return {
        'key': key,
        'weight': 1,
        'height': 0,
        'left': None,
        'right': None
    }


def comparer_(a, b):
    if a < b:
        return -1
    if a > b:
        return 1
    return 0


def height_(p):
    return p['height'] if p else 0


def weight_(p):
    return p['weight'] if p else 0


def b_factor_(p):
    return height_(p['right']) - height_(p['left'])


def count_height_and_weight_(p):
    hl = height_(p['left'])
    hr = height_(p['right'])
    p['height'] = (hl if hl > hr else hr) + 1

    wl = weight_(p['left'])
    wr = weight_(p['right'])
    p['weight'] = wl + wr + 1


def rotate_right_(p):
    q = p['left']
    p['left'] = q['right']
    q['right'] = p
    count_height_and_weight_(p)
    count_height_and_weight_(q)
    return q


def rotate_left_(q):
    p = q['right']
    q['right'] = p['left']
    p['left'] = q
    count_height_and_weight_(q)
    count_height_and_weight_(p)
    return p


def balance_(p):
    count_height_and_weight_(p)
    if b_factor_(p) == 2:
        if b_factor_(p['right']) < 0:
            p['right'] = rotate_right_(p['right'])
        return rotate_left_(p)
    if b_factor_(p) == -2:
        if b_factor_(p['left']) > 0:
            p['left'] = rotate_left_(p['left'])
        return rotate_right_(p)
    return p


def at_(p, k):
    if not p:
        return None
    wl = weight_(p['left'])
    if wl <= k < wl + 1:
        return p['key']
    elif k < wl:
        return at_(p['left'], k)
    else:
        return at_(p['right'], k - wl - 1)


def get_minimum_(p):
    if not p:
        return None
    return get_minimum_(p['left']) if p['left'] else p


def get_maximum_(p):
    if not p:
        return None
    return get_maximum_(p['right']) if p['right'] else p


def remove_minimum_(p):
    if not p['left']:
        return p['right']
    p['left'] = remove_minimum_(p['left'])
    return balance_(p)


def to_array_(p):
    arr = []
    if p['left']:
        arr += to_array_(p['left'])
    arr.append(p['key'])
    if p['right']:
        arr += to_array_(p['right'])
    return arr


def to_value_array_(p):
    arr = []
    if p['left']:
        arr += to_array_(p['left'])
    arr.append(p['key']['data'])
    if p['right']:
        arr += to_array_(p['right'])
    return arr


def avl_tree(comparer=None):
    if comparer is None:
        comparer = comparer_

    def count_(p, k):
        return upper_bound_(p, k) - lower_bound_(p, k)

    def get_size():
        return weight_(avl['root'])

    def get_min():
        p = get_minimum_(avl['root'])
        if p:
            return p['key']
        return None

    def get_max():
        p = get_maximum_(avl['root'])
        if p:
            return p['key']
        return None

    def lower_bound(k):
        return avl['lowerBound_'](avl['root'], k)

    def lower_bound_(p, k):
        if not p:
            return 0
        cmp = avl['comparer_'](k, p['key'])
        if cmp <= 0:
            return avl['lowerBound_'](p['left'], k)
        elif cmp > 0:
            return weight_(p['left']) + avl['lower_bound_'](p['right'], k) + 1

    def upper_bound(k):
        return avl['upperBound_'](avl['root'])

    def upper_bound_(p, k):
        if not p:
            return 0
        cmp = avl['comparer_'](k, p['key'])
        if cmp < 0:
            return avl['upperBound_'](p['left'], k)
        elif cmp >= 0:
            return weight_(p['left']) + avl['upperBound_'](p['right'], k) + 1

    def count(k):
        return count_(avl['root'], k)

    def at(k):
        return at_(avl['root'], k)

    def insert(k):
        avl['root'] = avl['insert_'](avl['root'], k)

    def insert_(p, k):
        if not p:
            return create_new_node_(k)
        cmp = avl['comparer_'](k, p['key'])
        if cmp < 0:
            p['left'] = avl['insert_'](p['left'], k)
        elif cmp >= 0:
            p['right'] = avl['insert_'](p['right'], k)
        return balance_(p)

    def remove(k):
        avl['root'] = avl['remove_'](avl['root'], k)

    def remove_(p, k):
        if not p:
            return None
        cmp = avl['comparer_'](k, p['key'])
        if cmp < 0:
            p['left'] = avl['remove_'](p['left'], k)
        elif cmp > 0:
            p['right'] = avl['remove_'](p['right'], k)
        else:
            q = p['left']
            r = p['right']
            if not r:
                return q
            min = get_minimum_(r)
            min['right'] = remove_minimum_(r)
            min['left'] = q
            return balance_(min)
        return balance_(p)

    def remove_at(k):
        val = avl['at'](k)
        avl['root'] = avl['remove_'](avl['root'], val)

    def to_array():
        if avl['root'] is None:
            return []
        return to_array_(avl['root'])

    avl = {
        'root': None,
        'comparer_': comparer,
        'size': get_size,
        'min': get_min,
        'max': get_max,
        'lowerBound': lower_bound,
        'lowerBound_': lower_bound_,
        'upper_bound': upper_bound,
        'upperBound_': upper_bound_,
        'count': count,
        'at': at,
        'insert': insert,
        'insert_': insert_,
        'remove': remove,
        'remove_': remove_,
        'removeAt': remove_at,
        'toArray': to_array
    }
    return avl



from .httpClient import HttpClient
from .metaApi_client import MetaApiClient
import pytest
http_client = HttpClient()
api_client = MetaApiClient(http_client, 'token')


@pytest.fixture(autouse=True)
async def run_around_tests():
    global http_client
    http_client = HttpClient()
    global api_client
    api_client = MetaApiClient(http_client, 'token')
    yield


class TestMetaApiClient:
    @pytest.mark.asyncio
    async def test_return_account_token_type(self):
        """Should return account token type."""
        assert api_client._token_type == 'account'

    @pytest.mark.asyncio
    async def test_return_api_token_type(self):
        """Should return api token type."""
        api_client = MetaApiClient(http_client, 'header.payload.sign')
        assert api_client._token_type == 'api'

    @pytest.mark.asyncio
    async def test_check_token_not_jwt(self):
        """Should check that current token is not JWT."""
        assert api_client._is_not_jwt_token()

    @pytest.mark.asyncio
    async def test_check_token_not_account(self):
        """Should check that current token is not account token."""
        api_client = MetaApiClient(http_client, 'header.payload.sign')
        assert api_client._is_not_account_token()

    @pytest.mark.asyncio
    async def test_handle_no_access_exception_with_account_token(self):
        """Should handle no access exception with account token."""
        try:
            api_client._handle_no_access_exception('methodName')
        except Exception as err:
            assert err.__str__() == 'You can not invoke methodName method, because you have connected with account ' + \
                   'access token. Please use API access token from https://app.metaapi.cloud/token page to invoke ' + \
                   'this method.'

    @pytest.mark.asyncio
    async def test_handle_no_access_exception_with_api_token(self):
        """Should handle no access exception with api token."""
        api_client = MetaApiClient(http_client, 'header.payload.sign')
        try:
            api_client._handle_no_access_exception('methodName')
        except Exception as err:
            assert err.__str__() == 'You can not invoke methodName method, because you have connected with API ' + \
                'access token. Please use account access token to invoke this method.'



from ..clients.metaApi.synchronizationListener import SynchronizationListener
from .models import MetatraderAccountInformation, MetatraderPosition, MetatraderOrder, \
    MetatraderSymbolSpecification, MetatraderSymbolPrice
import functools
from typing import List, Dict, Optional
from typing_extensions import TypedDict
import asyncio
from datetime import datetime


class TerminalStateDict(TypedDict):
    connected: bool
    connectedToBroker: bool
    accountInformation: Optional[dict]
    positions: List[dict]
    orders: List[dict]
    specifications: List[dict]
    specificationsBySymbol: dict
    pricesBySymbol: dict
    completedOrders: dict
    removedPositions: dict
    ordersInitialized: bool
    positionsInitialized: bool
    lastUpdateTime: float


class TerminalState(SynchronizationListener):
    """Responsible for storing a local copy of remote terminal state."""

    def __init__(self):
        """Inits the instance of terminal state class"""
        super().__init__()
        self._stateByInstanceIndex = {}

    @property
    def connected(self) -> bool:
        """Returns true if MetaApi has connected to MetaTrader terminal.

        Returns:
            Whether MetaApi has connected to MetaTrader terminal.
        """
        return True in list(map(lambda instance: instance['connected'], self._stateByInstanceIndex.values()))

    @property
    def connected_to_broker(self) -> bool:
        """Returns true if MetaApi has connected to MetaTrader terminal and MetaTrader terminal is connected to broker

        Returns:
             Whether MetaApi has connected to MetaTrader terminal and MetaTrader terminal is connected to broker
        """
        return True in list(map(lambda instance: instance['connectedToBroker'], self._stateByInstanceIndex.values()))

    @property
    def account_information(self) -> MetatraderAccountInformation:
        """Returns a local copy of account information.

        Returns:
            Local copy of account information.
        """
        return self._get_best_state()['accountInformation']

    @property
    def positions(self) -> List[MetatraderPosition]:
        """Returns a local copy of MetaTrader positions opened.

        Returns:
            A local copy of MetaTrader positions opened.
        """
        return self._get_best_state()['positions']

    @property
    def orders(self) -> List[MetatraderOrder]:
        """Returns a local copy of MetaTrader orders opened.

        Returns:
            A local copy of MetaTrader orders opened.
        """
        return self._get_best_state()['orders']

    @property
    def specifications(self) -> List[MetatraderSymbolSpecification]:
        """Returns a local copy of symbol specifications available in MetaTrader trading terminal.

        Returns:
             A local copy of symbol specifications available in MetaTrader trading terminal.
        """
        return self._get_best_state()['specifications']

    def specification(self, symbol: str) -> MetatraderSymbolSpecification:
        """Returns MetaTrader symbol specification by symbol.

        Args:
            symbol: Symbol (e.g. currency pair or an index).

        Returns:
            MetatraderSymbolSpecification found or undefined if specification for a symbol is not found.
        """
        state = self._get_best_state()
        return state['specificationsBySymbol'][symbol] if \
            (symbol in state['specificationsBySymbol']) else None

    def price(self, symbol: str) -> MetatraderSymbolPrice:
        """Returns MetaTrader symbol price by symbol.

        Args:
            symbol: Symbol (e.g. currency pair or an index).

        Returns:
            MetatraderSymbolPrice found or undefined if price for a symbol is not found.
        """
        state = self._get_best_state()
        return state['pricesBySymbol'][symbol] if \
            (symbol in state['pricesBySymbol']) else None

    async def on_connected(self, instance_index: int, replicas: int):
        """Invoked when connection to MetaTrader terminal established.

        Args:
            instance_index: Index of an account instance connected.
            replicas: Number of account replicas launched.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        self._get_state(instance_index)['connected'] = True
        self._reset_disconnect_timer(instance_index)

    async def on_disconnected(self, instance_index: int):
        """Invoked when connection to MetaTrader terminal terminated.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
             A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        state['connected'] = False
        state['connectedToBroker'] = False

    def _reset_disconnect_timer(self, instance_index: int):
        async def disconnect():
            await asyncio.sleep(60)
            await self.on_disconnected(instance_index)

        if hasattr(self, '_status_timer'):
            self._status_timer.cancel()
        self._status_timer = asyncio.create_task(disconnect())

    async def on_broker_connection_status_changed(self, instance_index: int, connected: bool):
        """Invoked when broker connection status have changed.

        Args:
            instance_index: Index of an account instance connected.
            connected: Is MetaTrader terminal is connected to broker.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        self._get_state(instance_index)['connectedToBroker'] = connected
        self._reset_disconnect_timer(instance_index)

    async def on_synchronization_started(self, instance_index: int):
        """Invoked when MetaTrader terminal state synchronization is started.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        state['accountInformation'] = None
        state['positions'] = []
        state['orders'] = []
        state['specifications'] = []
        state['specificationsBySymbol'] = {}
        state['pricesBySymbol'] = {}
        state['completedOrders'] = {}
        state['removedPositions'] = {}
        state['ordersInitialized'] = False
        state['positionsInitialized'] = False

    async def on_account_information_updated(self, instance_index: int,
                                             account_information: MetatraderAccountInformation):
        """Invoked when MetaTrader position is updated.

        Args:
            instance_index: Index of an account instance connected.
            account_information: Updated MetaTrader position.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        self._get_state(instance_index)['accountInformation'] = account_information

    async def on_positions_replaced(self, instance_index: int, positions: List[MetatraderPosition]):
        """Invoked when the positions are replaced as a result of initial terminal state synchronization.

        Args:
            instance_index: Index of an account instance connected.
            positions: Updated array of positions.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        state['positions'] = positions
        state['removedPositions'] = {}
        state['positionsInitialized'] = True

    async def on_position_updated(self, instance_index: int, position: MetatraderPosition):
        """Invoked when MetaTrader position is updated.

        Args:
            instance_index: Index of an account instance connected.
            position: Updated MetaTrader position.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        is_exists = False
        for i in range(len(state['positions'])):
            if state['positions'][i]['id'] == position['id']:
                state['positions'][i] = position
                is_exists = True
                break
        if (not is_exists) and (position['id'] not in state['removedPositions']):
            state['positions'].append(position)

    async def on_position_removed(self, instance_index: int, position_id: str):
        """Invoked when MetaTrader position is removed.

        Args:
            instance_index: Index of an account instance connected.
            position_id: Removed MetaTrader position id.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        position = next((p for p in state['positions'] if p['id'] == position_id), None)
        if position is None:
            for key in state['removedPositions']:
                e = state['removedPositions'][key]
                if e + 5 * 60 < datetime.now().timestamp():
                    del state['removedPositions'][key]
            state['removedPositions'][position_id] = datetime.now().timestamp()
        else:
            state['positions'] = list(filter(lambda p: p['id'] != position_id, state['positions']))

    async def on_orders_replaced(self, instance_index: int, orders: List[MetatraderOrder]):
        """Invoked when the orders are replaced as a result of initial terminal state synchronization.

        Args:
            instance_index: Index of an account instance connected.
            orders: Updated array of orders.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        state['orders'] = orders
        state['completedOrders'] = {}
        state['ordersInitialized'] = True

    async def on_order_updated(self, instance_index: int, order: MetatraderOrder):
        """Invoked when MetaTrader order is updated.

        Args:
            instance_index: Index of an account instance connected.
            order: Updated MetaTrader order.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        is_exists = False
        for i in range(len(state['orders'])):
            if state['orders'][i]['id'] == order['id']:
                state['orders'][i] = order
                is_exists = True
                break
        if (not is_exists) and (order['id'] not in state['completedOrders']):
            state['orders'].append(order)

    async def on_order_completed(self, instance_index: int, order_id: str):
        """Invoked when MetaTrader order is completed (executed or canceled).

        Args:
            instance_index: Index of an account instance connected.
            order_id: Completed MetaTrader order id.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        order = next((p for p in state['orders'] if p['id'] == order_id), None)
        if order is None:
            for key in state['completedOrders']:
                e = state['completedOrders'][key]
                if e + 5 * 60 < datetime.now().timestamp():
                    del state['completedOrders'][key]
            state['completedOrders'][order_id] = datetime.now().timestamp()
        else:
            state['orders'] = list(filter(lambda o: o['id'] != order_id, state['orders']))

    async def on_symbol_specification_updated(self, instance_index: int, specification: MetatraderSymbolSpecification):
        """Invoked when a symbol specification was updated

        Args:
            instance_index: Index of an account instance connected.
            specification: Updated MetaTrader symbol specification.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        for i in range(len(state['specifications'])):
            if state['specifications'][i]['symbol'] == specification['symbol']:
                state['specifications'][i] = specification
                break
        else:
            state['specifications'].append(specification)
        state['specificationsBySymbol'][specification['symbol']] = specification

    async def on_symbol_prices_updated(self, instance_index: int, prices: List[MetatraderSymbolPrice],
                                       equity: float = None, margin: float = None, free_margin: float = None,
                                       margin_level: float = None):
        """Invoked when prices for several symbols were updated.

        Args:
            instance_index: Index of an account instance connected.
            prices: Updated MetaTrader symbol prices.
            equity: Account liquidation value.
            margin: Margin used.
            free_margin: Free margin.
            margin_level: Margin level calculated as % of equity/margin.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        state = self._get_state(instance_index)
        state['lastUpdateTime'] = max(map(lambda p: p['time'].timestamp(), prices)) if len(prices) else 0
        prices_initialized = False
        if prices:
            for price in prices:
                state['pricesBySymbol'][price['symbol']] = price
                positions = list(filter(lambda p: p['symbol'] == price['symbol'], state['positions']))
                other_positions = list(filter(lambda p: p['symbol'] != price['symbol'], state['positions']))
                orders = list(filter(lambda o: o['symbol'] == price['symbol'], state['orders']))
                prices_initialized = True
                for position in other_positions:
                    if position['symbol'] in state['pricesBySymbol']:
                        p = state['pricesBySymbol'][position['symbol']]
                        if 'unrealizedProfit' not in position:
                            self._update_position_profits(position, p)
                    else:
                        prices_initialized = False
                for position in positions:
                    self._update_position_profits(position, price)
                for order in orders:
                    order['currentPrice'] = price['ask'] if (order['type'] == 'ORDER_TYPE_BUY' or
                                                             order['type'] == 'ORDER_TYPE_BUY_LIMIT' or
                                                             order['type'] == 'ORDER_TYPE_BUY_STOP' or
                                                             order['type'] == 'ORDER_TYPE_BUY_STOP_LIMIT') else \
                        price['bid']
        if state['accountInformation']:
            if state['positionsInitialized'] and prices_initialized:
                state['accountInformation']['equity'] = \
                    state['accountInformation']['balance'] + functools.reduce(
                        lambda a, b: a + (b['unrealizedProfit'] if 'unrealizedProfit' in b else 0),
                        state['positions'], 0)
            else:
                state['accountInformation']['equity'] = equity if equity else (
                    state['accountInformation']['equity'] if 'equity' in state['accountInformation'] else None)
            state['accountInformation']['margin'] = margin if margin else (
                state['accountInformation']['margin'] if 'margin' in state['accountInformation'] else None)
            state['accountInformation']['freeMargin'] = free_margin if free_margin else (
                state['accountInformation']['freeMargin'] if 'freeMargin' in state['accountInformation'] else None)
            state['accountInformation']['marginLevel'] = margin_level if free_margin else (
                state['accountInformation']['marginLevel'] if 'marginLevel' in state['accountInformation'] else None)

    def _update_position_profits(self, position: Dict, price: Dict):
        specification = self.specification(position['symbol'])
        if specification:
            if 'unrealizedProfit' not in position or 'realizedProfit' not in position:
                position['unrealizedProfit'] = (1 if (position['type'] == 'POSITION_TYPE_BUY') else -1) * \
                                               (position['currentPrice'] - position['openPrice']) * \
                                               position['currentTickValue'] * position['volume'] / \
                                               specification['tickSize']
                position['realizedProfit'] = position['profit'] - position['unrealizedProfit']
            new_position_price = price['bid'] if (position['type'] == 'POSITION_TYPE_BUY') else price['ask']
            is_profitable = (1 if (position['type'] == 'POSITION_TYPE_BUY') else -1) * (new_position_price -
                                                                                        position['openPrice'])
            current_tick_value = price['profitTickValue'] if (is_profitable > 0) else price['lossTickValue']
            unrealized_profit = (1 if (position['type'] == 'POSITION_TYPE_BUY') else -1) * \
                                (new_position_price - position['openPrice']) * current_tick_value * \
                position['volume'] / specification['tickSize']
            position['unrealizedProfit'] = unrealized_profit
            position['profit'] = position['unrealizedProfit'] + position['realizedProfit']
            position['currentPrice'] = new_position_price
            position['currentTickValue'] = current_tick_value

    def _get_state(self, instance_index: int) -> TerminalStateDict:
        if str(instance_index) not in self._stateByInstanceIndex:
            self._stateByInstanceIndex[str(instance_index)] = self._construct_terminal_state()
        return self._stateByInstanceIndex[str(instance_index)]

    def _construct_terminal_state(self) -> TerminalStateDict:
        return {
            'connected': False,
            'connectedToBroker': False,
            'accountInformation': None,
            'positions': [],
            'orders': [],
            'specifications': [],
            'specificationsBySymbol': {},
            'pricesBySymbol': {},
            'completedOrders': {},
            'removedPositions': {},
            'ordersInitialized': False,
            'positionsInitialized': False,
            'lastUpdateTime': 0
        }

    def _get_best_state(self) -> TerminalStateDict:
        result = None
        max_update_time = None
        for state in self._stateByInstanceIndex.values():
            if max_update_time is None or max_update_time < state['lastUpdateTime']:
                max_update_time = state['lastUpdateTime']
                result = state
        return result or self._construct_terminal_state()



from typing_extensions import TypedDict
from httpx import Response
from ..metaApi_client import MetaApiClient
from typing import List


class ProvisioningProfileDto(TypedDict):
    """Provisioning profile model"""

    _id: str
    """Provisioning profile unique identifier"""
    name: str
    """Provisioning profile name"""
    version: int
    """MetaTrader version (allowed values are 4 and 5)"""
    status: str
    """Provisioning profile status (allowed values are new and active)"""
    brokerTimezone: str
    """Broker timezone name from Time Zone Database."""
    brokerDSTSwitchTimezone: str
    """Broker DST switch timezone name from Time Zone Database."""


class NewProvisioningProfileDto(TypedDict):
    """New provisioning profile model."""

    name: str
    """Provisioning profile name."""
    version: int
    """MetaTrader version (allowed values are 4 and 5)."""
    brokerTimezone: str
    """Broker timezone name from Time Zone Database."""
    brokerDSTSwitchTimezone: str
    """Broker DST switch timezone name from Time Zone Database."""


class ProvisioningProfileIdDto(TypedDict):
    """Provisioning profile id model."""

    id: str
    """Provisioning profile unique identifier."""


class ProvisioningProfileUpdateDto(TypedDict):
    """Updated provisioning profile data."""

    name: str
    """Provisioning profile name."""


class ProvisioningProfileClient(MetaApiClient):
    """metaapi.cloud provisioning profile API client (see https://metaapi.cloud/docs/provisioning/)"""

    async def get_provisioning_profiles(self, version: int, status: str) -> 'Response[List[ProvisioningProfileDto]]':
        """Retrieves provisioning profiles owned by user
        (see https://metaapi.cloud/docs/provisioning/api/provisioningProfile/readProvisioningProfiles/)

        Args:
            version: Optional version filter (allowed values are 4 and 5).
            status: Optional status filter (allowed values are new and active).

        Returns:
            A coroutine resolving with provisioning profiles found
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_provisioning_profiles')
        params = {}
        if version:
            params['version'] = version
        if status:
            params['status'] = status
        opts = {
            'url': f"{self._host}/users/current/provisioning-profiles",
            'method': 'GET',
            'params': params,
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def get_provisioning_profile(self, id: str) -> 'Response[ProvisioningProfileDto]':
        """Retrieves a provisioning profile by id (see
        https://metaapi.cloud/docs/provisioning/api/provisioningProfile/readProvisioningProfile/). Throws an error if
        profile is not found.

        Args:
            id: Provisioning profile id.

        Returns:
            A coroutine resolving with provisioning profile found.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('get_provisioning_profile')
        opts = {
            'url': f"{self._host}/users/current/provisioning-profiles/{id}",
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def create_provisioning_profile(self, provisioning_profile: NewProvisioningProfileDto) -> Response:
        """Creates a new provisioning profile (see
        https://metaapi.cloud/docs/provisioning/api/provisioningProfile/createNewProvisioningProfile/). After creating
        a provisioning profile you are required to upload extra files in order to activate the profile for further use.

        Args:
            provisioning_profile: Provisioning profile to create.

        Returns:
            A coroutine resolving with an id of the provisioning profile created.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('create_provisioning_profile')
        opts = {
            'url': f"{self._host}/users/current/provisioning-profiles",
            'method': 'POST',
            'headers': {
                'auth-token': self._token
            },
            'body': provisioning_profile
        }
        return await self._httpClient.request(opts)

    async def upload_provisioning_profile_file(self, provisioning_profile_id: str,
                                               file_name: str, file: str or memoryview) -> Response:
        """Uploads a file to a provisioning profile (see
        https://metaapi.cloud/docs/provisioning/api/provisioningProfile/uploadFilesToProvisioningProfile/).

        Args:
            provisioning_profile_id: Provisioning profile id to upload file to.
            file_name: Name of the file to upload. Allowed values are servers.dat for MT5 profile, broker.srv for
            MT4 profile.
            file: Path to a file to upload or buffer containing file contents.

        Returns:
            A coroutine resolving when file upload is completed.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('upload_provisioning_profile_file')
        if type(file) == str:
            file = open(file, 'rb').read()
        opts = {
            'method': 'PUT',
            'url': f'{self._host}/users/current/provisioning-profiles/{provisioning_profile_id}/{file_name}',
            'files': {
                'file': file
            },
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def delete_provisioning_profile(self, id: str) -> Response:
        """ Deletes a provisioning profile (see
        https://metaapi.cloud/docs/provisioning/api/provisioningProfile/deleteProvisioningProfile/).
        Please note that in order to delete a provisioning profile you need to delete MT accounts connected to it first.

        Args:
            id: Provisioning profile id.

        Returns:
            A coroutine resolving when provisioning profile is deleted.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('delete_provisioning_profile')
        opts = {
            'url': f'{self._host}/users/current/provisioning-profiles/{id}',
            'method': 'DELETE',
            'headers': {
                'auth-token': self._token
            }
        }
        return await self._httpClient.request(opts)

    async def update_provisioning_profile(self, id: str, provisioning_profile: ProvisioningProfileUpdateDto):
        """Updates existing provisioning profile data (see
        https://metaapi.cloud/docs/provisioning/api/provisioningProfile/updateProvisioningProfile/).

        Args:
            id: Provisioning profile id.
            provisioning_profile: Updated provisioning profile.

        Returns:
            A coroutine resolving when provisioning profile is updated.
        """
        if self._is_not_jwt_token():
            return self._handle_no_access_exception('update_provisioning_profile')
        opts = {
            'url': f'{self._host}/users/current/provisioning-profiles/{id}',
            'method': 'PUT',
            'headers': {
                'auth-token': self._token
            },
            'body': provisioning_profile
        }
        return await self._httpClient.request(opts)



from .models import MetatraderDeal, MetatraderOrder
from typing import List
from .memoryHistoryStorageModel import MemoryHistoryStorageModel
from .historyFileManager import HistoryFileManager
from datetime import datetime
from .models import date
import pytz


class MemoryHistoryStorage(MemoryHistoryStorageModel):
    """History storage which stores MetaTrader history in RAM."""

    def __init__(self, account_id: str, application: str = 'MetaApi'):
        """Inits the in-memory history store instance"""
        super().__init__()
        self._accountId = account_id
        self._fileManager = HistoryFileManager(account_id, application, self)
        self._deals = []
        self._historyOrders = []
        self._lastDealTimeByInstanceIndex = {}
        self._lastHistoryOrderTimeByInstanceIndex = {}
        self._fileManager.start_update_job()

    async def initialize(self):
        """Initializes the storage and loads required data from a persistent storage."""
        await self.load_data_from_disk()

    @property
    def deals(self) -> List[MetatraderDeal]:
        """Returns all deals stored in history storage.

        Returns:
            All deals stored in history storage.
        """
        return self._deals

    @property
    def history_orders(self) -> List[MetatraderOrder]:
        """Returns all history orders stored in history storage.

        Returns:
            All history orders stored in history storage.
        """
        return self._historyOrders

    @property
    def last_deal_time_by_instance_index(self) -> dict:
        """Returns times of last deals by instance indices.

        Returns:
            Dictionary of last deal times by instance indices."""
        return self._lastDealTimeByInstanceIndex

    @property
    def last_history_order_time_by_instance_index(self) -> dict:
        """Returns times of last history orders by instance indices.

        Returns:
            Dictionary of last history orders times by instance indices."""
        return self._lastHistoryOrderTimeByInstanceIndex

    async def clear(self):
        """Clears the storage and deletes persistent data."""
        self._deals = []
        self._historyOrders = []
        self._lastDealTimeByInstanceIndex = {}
        self._lastHistoryOrderTimeByInstanceIndex = {}
        await self._fileManager.delete_storage_from_disk()

    async def load_data_from_disk(self):
        """Loads history data from the file manager.

        Returns:
            A coroutine which resolves when the history is loaded.
        """
        history = await self._fileManager.get_history_from_disk()
        self._deals = history['deals']
        self._historyOrders = history['historyOrders']
        self._lastDealTimeByInstanceIndex = history['lastDealTimeByInstanceIndex'] if 'lastDealTimeByInstanceIndex' \
                                                                                      in history else {}
        self._lastHistoryOrderTimeByInstanceIndex = history['lastHistoryOrderTimeByInstanceIndex'] if \
            'lastHistoryOrderTimeByInstanceIndex' in history else {}

    async def update_disk_storage(self):
        """Saves unsaved history items to disk storage.

        Returns:
            A coroutine which resolves when the history is saved.
        """
        await self._fileManager.update_disk_storage()

    async def last_history_order_time(self, instance_index: int = None) -> datetime:
        """Returns the time of the last history order record stored in the history storage.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
            The time of the last history order record stored in the history storage.
        """
        if instance_index is not None:
            return date(self._lastHistoryOrderTimeByInstanceIndex[str(instance_index)] if str(instance_index) in
                        self._lastHistoryOrderTimeByInstanceIndex else 0)
        else:
            return date(max(list(self._lastHistoryOrderTimeByInstanceIndex.values())))

    async def last_deal_time(self, instance_index: int = None) -> datetime:
        """Returns the time of the last history deal record stored in the history storage.

        Args:
            instance_index: Index of an account instance connected.

        Returns:
            The time of the last history deal record stored in the history storage.
        """
        if instance_index is not None:
            return date(self._lastDealTimeByInstanceIndex[str(instance_index)] if str(instance_index) in
                        self._lastDealTimeByInstanceIndex else 0)
        else:
            return date(max(list(self._lastDealTimeByInstanceIndex.values())))

    async def on_history_order_added(self, instance_index: int, history_order: MetatraderOrder):
        """Invoked when a new MetaTrader history order is added.

        Args:
            instance_index: Index of an account instance connected.
            history_order: New MetaTrader history order.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        insert_index = 0
        replacement_index = -1

        def get_done_time(order):
            if 'doneTime' in order:
                return order['doneTime'].timestamp() if (isinstance(order['doneTime'], datetime)) \
                    else date(order['doneTime']).timestamp()
            else:
                return 0

        new_history_order_time = get_done_time(history_order)
        if str(instance_index) not in self._lastHistoryOrderTimeByInstanceIndex or \
                self._lastHistoryOrderTimeByInstanceIndex[str(instance_index)] < new_history_order_time:
            self._lastHistoryOrderTimeByInstanceIndex[str(instance_index)] = new_history_order_time

        for i in range(len(self._historyOrders)):
            index = len(self._historyOrders) - 1 - i
            order = self._historyOrders[index]
            order_time = get_done_time(order)
            if (order_time < new_history_order_time) or \
               (order_time == new_history_order_time and order['id'] <= history_order['id']):
                if (order_time == new_history_order_time and order['id'] == history_order['id'] and
                   order['type'] == history_order['type']):
                    replacement_index = index
                else:
                    insert_index = index + 1
                break
        if replacement_index != -1:
            self._historyOrders[replacement_index] = history_order
            self._fileManager.set_start_new_order_index(replacement_index)
        else:
            self._historyOrders.insert(insert_index, history_order)
            self._fileManager.set_start_new_order_index(insert_index)

    async def on_deal_added(self, instance_index: int, new_deal: MetatraderDeal):
        """Invoked when a new MetaTrader history deal is added.

        Args:
            instance_index: Index of an account instance connected.
            new_deal: New MetaTrader history deal.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        insert_index = 0
        replacement_index = -1

        def get_time(deal):
            if 'time' in deal:
                return deal['time'].timestamp() if (isinstance(deal['time'], datetime)) \
                    else date(deal['time']).timestamp()
            else:
                return 0

        new_deal_time = get_time(new_deal)
        if str(instance_index) not in self._lastDealTimeByInstanceIndex or \
                self._lastDealTimeByInstanceIndex[str(instance_index)] < new_deal_time:
            self._lastDealTimeByInstanceIndex[str(instance_index)] = new_deal_time
        for i in range(len(self._deals)):
            index = len(self._deals) - 1 - i
            deal = self._deals[index]
            deal_time = get_time(deal)
            if (deal_time < new_deal_time) or \
                    (deal_time == new_deal_time and deal['id'] <= new_deal['id']):
                if (deal_time == new_deal_time and deal['id'] == new_deal['id'] and
                        deal['type'] == new_deal['type']):
                    replacement_index = index
                else:
                    insert_index = index + 1
                break
        if replacement_index != -1:
            self._deals[replacement_index] = new_deal
            self._fileManager.set_start_new_deal_index(replacement_index)
        else:
            self._deals.insert(insert_index, new_deal)
            self._fileManager.set_start_new_deal_index(insert_index)

    async def on_deal_synchronization_finished(self, instance_index: int, synchronization_id: str):
        """Invoked when a synchronization of history deals on a MetaTrader account have finished.

        Args:
            instance_index: Index of an account instance connected.
            synchronization_id: Synchronization request id.

        Returns:
            A coroutine which resolves when the asynchronous event is processed.
        """
        self._dealSynchronizationFinished[str(instance_index)] = True
        await self.update_disk_storage()
